/** 
 * CustomAlloc.h - Custom allocator using dmalloc
 * 
 * Copyright (c) 2006, Amanjit Singh Gill
 * All rights reserved.
 * 
 * Redistribution and use in source and binary  forms, with or without
 * modification, are permitted provided  that the following conditions
 * are met:
 *
 * Redistributions of  source  code must  retain  the  above copyright
 * notice, this  list  of conditions   and the  following  disclaimer.
 * Redistributions in  binary form must  reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or   other     materials   provided  with     the
 * distribution. Neither the name of Amanjtit Singh Gill nor the names
 * of its contributors   may be used  to  endorse or promote  products
 * derived      from this software      without specific prior written
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND ANY EXPRESS OR  IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS
 * FOR  A PARTICULAR PURPOSE  ARE DISCLAIMED.  IN  NO  EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,   SPECIAL,   EXEMPLARY,     OR CONSEQUENTIAL    DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR  PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR TORT   (INCLUDING  NEGLIGENCE OR  OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * Based on Custom allocator sample code for vector
 * http://www.josuttis.com/libbook/memory/CustomAlloc.hpp.html
 */

#ifndef contprj_customalloc
#define contprj_customalloc

#include <limits>
#include <iostream>
#include <iomanip>
#include <new>
#include <vector>
#include <string>
#include <sstream>
#include <cassert>
#include <stdio.h>
#include <cassert>
#include "dmalloc.h"
#ifdef _WIN32
# include "crtdbg.h"
#endif

using namespace std;

const size_t MINSIZE = BUFSIZ;

#ifdef _WIN32
# ifdef _DEBUG
#   if 0
#     define CALLOC_DEBUG
#   endif
# else
#   undef CALLOC_DEBUG
# endif
#endif

#ifdef CALLOC_DEBUG
  #define CERR cerr << "[" << setw (12) << std::left << __FILE__                            \
                    << ", #" << setw (3) << setfill (' ') << std::right << __LINE__ << "] "
  #define CHECK_PRINT(x) CERR << "In function ---> " << x << endl
#else
  #define CERR
  #define CHECK_PRINT(x)
#endif

/**
 * The Custom Allocator
 */
template <class TYPE_>
class CustomAlloc {
  public:
    // type definitions
    typedef TYPE_          value_type;
    typedef TYPE_*         pointer;
    typedef const TYPE_*   const_pointer;
    typedef TYPE_&         reference;
    typedef const TYPE_&   const_reference;
    typedef size_t         size_type;
    typedef std::ptrdiff_t difference_type;

    /** rebind allocator to type OTHER_ */
    template <class OTHER_>
    struct rebind {
      typedef CustomAlloc<OTHER_> other;
    };

    /** return address of values  */
    inline pointer address (reference value) const {
      CHECK_PRINT ("address (reference)");
      return &value;
    }

    inline const_pointer address (const_reference value) const {
      CHECK_PRINT ("address (const_reference)");
      return &value;
    }

    /*
     * constructors and destructor
     * nothing to do because the allocator has no state
     */
    CustomAlloc() throw() {
      CHECK_PRINT ("CustomAlloc()");
    }

    CustomAlloc(const CustomAlloc&) throw() {
      CHECK_PRINT ("CustomAlloc(const CustomAlloc&)");
    }

    template <class OTHER_>
    CustomAlloc (const CustomAlloc<OTHER_>&) throw() {
      CHECK_PRINT ("CustomAlloc(const CustomAlloc<OTHER_>&)");
    }

    ~CustomAlloc() throw() {
      CHECK_PRINT ("~CustomAlloc()");
    }

    /** return maximum number of elements that can be allocated */
    inline size_type max_size () const throw() {
      CHECK_PRINT ("max_size () const");
      return numeric_limits<size_t>::max() / sizeof(TYPE_);
    }

    /** allocate but don't initialize num elements of type TYPE_ */
    inline pointer allocate (size_type num, const void* = 0) {
      CHECK_PRINT ("allocate (size_type, const void*)");
      #ifdef CALLOC_DEBUG
      CERR << "allocate " << (unsigned int)num << " element(s)" << " of size " << sizeof(TYPE_);
      #endif
      pointer ret (NULL);
      try {
        ret = static_cast<pointer>(::dlmalloc(num*sizeof(TYPE_)));
      }
      catch(bad_alloc&) {
        #ifdef CALLOC_DEBUG
        CERR <<"Memory allocation failure in allocate()!" << endl;
        #endif
        exit(1);
      }
      #ifdef CALLOC_DEBUG
      cerr << ";  allocated at: " << (void*)ret << endl;
      #endif
      return ret;
    }

    /** initialize elements of allocated storage p with value value */
    inline void construct (pointer p, const TYPE_& value) {
      CHECK_PRINT ("construct (pointer, const TYPE_&)");
      // initialize memory with placement new
      try {
          new(static_cast<void*>(p))TYPE_(value);
      }
      catch(bad_alloc&) {
      #ifdef CALLOC_DEBUG
      CERR <<"Memory allocation failure in construct()!" << endl; exit(1);
      #endif
      }
    }

    /** destroy elements of initialized storage p */
    inline void destroy (pointer p) {
      CHECK_PRINT ("destroy (pointer)");
      // destroy objects by calling their destructor
      p->~TYPE_();
    }

    /** deallocate storage p of deleted elements */
    inline void deallocate (pointer p, size_type num) {
      CHECK_PRINT ("deallocate (pointer, size_type)");
      // print message and deallocate memory with global delete
      #ifdef CALLOC_DEBUG
        CERR << "deallocate " << (unsigned int) num << " element(s)"
            << " of size " << sizeof(TYPE_)
            << " at: " << static_cast<void*>(p) << endl;
      #endif
      return ::dlfree(p);
    }
};

// return that all specializations of this allocator are interchangeable
template <class TYPE1_, class TYPE2_>
bool operator== (const CustomAlloc<TYPE1_>&, const CustomAlloc<TYPE2_>&) throw() {
  CHECK_PRINT ("operator== (const CustomAlloc<TYPE1_>&, const CustomAlloc<TYPE2_>&)");
  return true;
}

template <class TYPE1_, class TYPE2_>
bool operator!= (const CustomAlloc<TYPE1_>&, const CustomAlloc<TYPE2_>&) throw() {
  CHECK_PRINT ("operator!= (const CustomAlloc<TYPE1_>&, const CustomAlloc<TYPE2_>&)");
  return false;
}

#endif
